/* $RCSfile: $
 * $Revision: $ $Date: $
 * Auth: Jochen Fritz (jfritz@steptools.com)
 * 
 * Copyright (c) 1991-2013 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 */

"use strict";



P21.readModel = function(url, callbacks)
{
//    console.log ("Reading model: "+url);

    
    var model = new P21.Model(url);
    
    var parse = new P21.Parser();

    if (callbacks) {
	for (var ev in callbacks) {
	    (function(ev_name, ev_fn) {
	    
		parse[ev_name] = function() {
		    var args = Array.prototype.slice.call(arguments);		
		    
		
		    args.unshift(model);
		    ev_fn.apply(null, args);
		}
	    })(ev, callbacks[ev]);
	}
    }

    if (callbacks && callbacks.init)
	callbacks.init(model);

    parse.beginDataSection = function() {
	if (!callbacks)
	    throw new P21.Lexer.CancelParse();
    }

    parse.anchor = function(name, value, tags) {
	model.anchor[name] = value;
	
//	console.log ("Have anchor: "+name+"="+P21.toP21String(value));

	var mtag = {};
	model.anchor.$tag[name] = mtag;
	
	for (var i=0; i<tags.length; i++) {
	    var t = tags[i];
	    mtag[t.name] = t.value;
	}
    }

    parse.reference = function(name, value, tags) {
	name = name.toString();
	
	if (name.match(/^[\d_]/))
	    name = '_' + name;
	    model.reference[name] = value;

	var mtag = {};
	model.reference.$tag[name] = mtag;
	    
	for (var i=0; i<tags.length; i++) {
	    var t = tags[i];
	    mtag[t.name] = t.value;
	}
    }

    
    /////////////////////////////////////
    
    var lexer = new P21.Lexer();
    lexer.processToken = function(tok) {
	parse.token(tok);
    }
    lexer.readFile(url);    

    return model;
    
}


P21.copyUpdateModel = function(out_url, model)
{
    function write_token(out_fd, tok) {
	
	if (tok.type == 'comment')
	    return;
	
	var str = P21.toP21String(tok.value);
	write_file(out_fd, str);
	
	if (tok.type == 'punct' && tok.value == ';')
	    write_file(out_fd, "\n");
    }

    function write_anchor(out_fd, model) {
	var sect = model.anchor;
	var found = false;

	var kw;
	for (kw in sect) {
	    if (kw == '$tag') continue;

	    found = true;
	    break;
	}

	if (!found)
	    return;
	
	write_file(out_fd, "ANCHOR;\n");

	var tags = sect.$tag;
	
	for (kw in sect) {
	    if (kw == "$tag") continue;
	    
	    write_file(out_fd, "<"+kw+">");
	    var tags = sect.$tag[kw];
	    for (var tag in tags) {
		write_file(out_fd, " {"+tag+": "+P21.toP21String(tags[tag])
			   + "}");
	    }

	    write_file (out_fd, " = " + P21.toP21String(sect[kw]) + ";\n");
	}
	
	write_file(out_fd, "ENDSEC;\n");
    }

    function get_timestamp() {
    
	function pad(number) {
	    var r = String(number);
	    if ( r.length === 1 ) {
		r = '0' + r;
	    }
	    return r;
	}
	
	function get_tz(off) {

	    var prefix;
	    
	    if (off >= 0) {
		prefix = '+';
	    }
	    else {
		prefix = '-';
		off = -off;
	    }
	    
	    var hour = Math.floor(off / 60);
	    var min = off % 60;
	    
	    return prefix + hour + ':' + pad(min);
	}

	function toISODate(date) {
	    return date.getFullYear()
		+ '-' + pad( date.getMonth() + 1 )
		+ '-' + pad( date.getDate() )
		+ 'T' + pad( date.getHours() )
		+ ':' + pad( date.getMinutes() )
		+ ':' + pad( date.getSeconds() )
		+ get_tz(date.getTimezoneOffset());
	};


	var now = new Date();
	return toISODate(now);	
    }

    
    function write_reference(out_fd, model) {
	var sect = model.reference;
	var found = false;

	var kw;
	for (kw in sect) {
	    if (kw == '$tag') continue;

	    found = true;
	    break;
	}

	if (!found)
	    return;
	
	write_file(out_fd, "REFERENCE;\n");

	var tags = sect.$tag;
	
	for (kw in sect) {
	    if (kw == "$tag") continue;

	    var name = kw;
	    if (name.charAt(0) == '_')
		name = name.substr(1);
	    
	    write_file(out_fd, "#"+name);
	    var tags = sect.$tag[kw];
	    for (var tag in tags) {
		write_file(out_fd, " {"+tag+": "+P21.toP21String(tags[tag])
			   + "}");
	    }

	    write_file (out_fd, " = " + P21.toP21String(sect[kw]) + ";\n");
	}
	
	write_file(out_fd, "ENDSEC;\n");
    }

    console.log ("Wrtiting to file:" +out_url);
    var out_fd = open_file_write(out_url);
    
    // Do we copy the source to the destination
    var copy = true;
    var copy_after = false;

    var continue_parse = true;
    var anchor_written = false;
    
    var parse = new P21.Parser();


    parse.beginHeaderSection = function() {
	copy = false;
	write_file(out_fd, "HEADER;\n");
    }
    parse.endHeaderSection = function() {
	copy_after = true;
	write_file(out_fd, "ENDSECT;\n");
    }
    
    parse.headerEntity = function(name, vals) {
	if (name == 'FILE_NAME')
	    vals[1] = new P21.String(get_timestamp());
	
	write_file(out_fd, name + " "+P21.toP21String(vals) +";\n");
//	console.log ("Have header instance: "+name);
//	console.log ("  = "+P21.toP21String(vals));
    }
    
    
    parse.beginAnchorSection = function() {copy = false;   }
    parse.endAnchorSection = function() {copy_after = true;}

    parse.beginReferenceSection = function() {copy = false;}
    parse.endReferenceSection = function() {copy_after = true;}

    parse.beginDataSection = function() {
	continue_parse = false;
	write_anchor(out_fd, model);
	write_reference(out_fd, model);
    }

    parse.endFile = function() {
	continue_parse = false;
	write_anchor(out_fd, model);
	write_reference(out_fd, model);
    }

    
    
    /////////////////////////////////////
    
    var lexer = new P21.Lexer();
    lexer.processToken = function(tok) {
	if (continue_parse)
	    parse.token(tok);
	
	if (copy)
	    write_token(out_fd, tok);

	if (copy_after) {
	    copy_after = false;
	    copy = true;
	}
	
    }
    lexer.readFile(model.url);    

    close_file_write(out_fd);
    
    return model;
    
}
