/* $RCSfile: $
 * $Revision: $ $Date: $
 * Auth: Jochen Fritz (jfritz@steptools.com)
 * 
 * 	Copyright (c) 1991-2013 by 
 * 	STEP Tools Inc., Troy, New York
 * 	All Rights Reserved
 * 
 * 	This software is furnished under a license and may be used and
 * 	copied only in accordance with the terms of such license and with
 * 	the inclusion of the above copyright notice.  This software and
 * 	accompanying written materials or any other copies thereof may
 * 	not be provided or otherwise made available to any other person.
 * 	No title to or ownership of the software is hereby transferred.
 * 
 * 		----------------------------------------
 */

var P21 = {};

function SUBTYPE(base) {
    function _subclassOf() {};
    _subclassOf.prototype= base.prototype;
    var ret = new _subclassOf();
    return ret;
}



P21.Wrapper = (function() {

    function cls() {};

    cls.prototype.toP21String = function()
    {
	throw new Error("toP21String must be implemented for subscalss ");
    }

    return cls;
})();


P21.toP21String = function(val) {
    if (val == null)
	return '$';
    
    else if (val === true) {
	return '.T.';
    }

    else if (val === false) {
	return '.F.';
    }

    else if (val.toP21String)
	return val.toP21String();

    else if (val instanceof Array) {
	var ret = "(";
	for (var i=0; i<val.length; i++) {
	    if (i > 0)
		ret += ", ";
	    ret += P21.toP21String(val[i]);
	}
	ret += ")";
	return ret;
    }
    
    return val.toString();
}
    


P21.Integer = (function() {

    function cls(val) {
	if (val == null)
	    return null;

	val = val.valueOf();

	if (typeof val === 'string') {
	    if (! /^\d+$/.test(val))
		throw new Error ("Integer syntax error: "+val);
	    
	    val = parseInt(val, 10);
	}
	
	if (typeof val !== 'number')
	    throw new Error("Parameter is not a number");
		
	this.value = val;
    }
    
    cls.prototype = SUBTYPE(P21.Wrapper);

    cls.prototype.toString = function()
    {
	return new Number(this.value).toString();
    }
    
    cls.prototype.valueOf = function()
    {
	return this.value;
    }
    
    cls.prototype.toP21String = function()
    {
	return this.toString();
    }
       
    return cls;
})();


P21.Real = (function() {

    function cls(val) {
	if (val == null)
	    return null;
	
	val = val.valueOf();

	if (typeof val === 'string') {
	    if (! /^[+-]?\d+\.(\d*)?([eE][+-]?\d*)?$/.test(val))
		throw new Error ("Real syntax error");
	    
	    val = parseFloat(val);
	}
	
	if (typeof val !== 'number')
	    throw new Error("Parameter is not a number");
		
	this.value = val;
    }
    
    cls.prototype = SUBTYPE(P21.Wrapper);

    cls.prototype.toString = function()
    {
	return new Number(this.value).toString();
    }
    
    cls.prototype.valueOf = function()
    {
	return this.value;
    }

    
    cls.prototype.toP21String = function()
    {
	var str = this.toString();

	str = str.replace(/\+/, "");
	
	if (str.indexOf('.') >= 0)
	    return str;
	
	if (str.search(/[eE]/)) {
	    return str.replace(/[eE]/, ".E");
	}

	return str + ".";
    }
       
    return cls;
})();



P21.String = (function() {

    var cls = function(val) {
	if (val == null)
	    return null;
	
 	val = val.valueOf();
	
 	if (typeof val !== 'string') {
 	    val = new String(val).toString()
 	}
	
 	this.value = val;
    }
    
    cls.encodeP21String = function(str, use_utf8) {
	var ret = "";
     	var len = str.length;
	
	for (var i=0; i<len; i++) {
	    var ch = str.charAt(i);
	    
	    if (ch == ' ' ||
		ch >= '0' && ch <= '9' ||
		ch >= 'a' && ch <= 'z' ||
		ch >= 'A' && ch <= 'Z' || ch == '_' ||
		ch == '!' || ch == '"' || ch == '*' || ch == '$' ||
		ch == '%' || ch == '&' || ch == '.' || ch == '#' ||
		ch == '+' || ch == ',' || ch == '-' || ch == '(' ||
		ch == ')' || ch == '?' || ch == '/' || ch == ':' ||
		ch == '<' || ch == '=' || ch == '>' || ch == ';' ||
		ch == '@' || ch == '[' || ch == ']' || ch == '{' ||
		ch == '|' || ch == '}' || ch == '^' || ch == '`' ||
		ch == '~')
    		ret += ch;
	    
    	    else if (ch == "'")
    		ret += "''";
	    else if (ch == '\\')
		ret += "\\\\";

	    else {
		var code = ch.charCodeAt(0);

		if (code <= 0xff) {
		    // 8-bit character
		    var hex = ("00" + code.toString(16).toUpperCase())
			.substr(-2);
    		    ret += "\\X\\" + hex;
		}

		else if (code < 0xd800 || code > 0xdfff) {
		    // 16 bit character
		    var hex = ("0000" + code.toString(16).toUpperCase())
			.substr(-4);
		    
		    ret += "\\X2\\" + hex + "\\X0\\";		    
		}

		else if (0xd800 <= code && code <= 0xDBFF) {
		    // High surrogate 
		    if (len <= (i+1))  {
			throw new Error('High surrogate without following '
					+'low surrogate');
		    }
		    i++;
		    var next = str.charCodeAt(i);
		    if (0xDC00 > next || next > 0xDFFF) {
			throw new Error('High surrogate without following '
					+'low surrogate');
		    }


		    var ls = code;  // Lead surrogate
		    var ts = next;  // trail surrogate

		    var c_hi = ls & 0x3ff;
		    var c_lo = ts & 0x3ff;

		    console.log ("Hi="+c_hi+ " "+"lo="+c_lo);
		    
		    code = ((c_hi << 10) | c_lo) + 0x10000;

		    var hex = ("00000000" + code.toString(16).toUpperCase())
			.substr(-8);
		    
		    ret += "\\X4\\" + hex + "\\X0\\";		    
		    
    		}

		else {
		    throw new Error ("Unexpected char code: "+code);
		}
	    }
	}
    	return ret;
    }

    
    function assert_hex(ch) {
	if (ch >= '0' && ch <= '9')
	    return;
	if (ch >= 'A' && ch <= 'F')
	    return;

	throw new Error ("Expecting hex digit");
    }

    
    cls.decodeP21String = function(str) {
	var ret = "";

	var state = 0;
	var hex;
	var hex_sz;      // nummber of hex digits we are expecting
	
	for (var i=0; i<str.length; i++) {
	    var ch = str.charAt(i);

	    switch (state) {

	    case 0:
		if (ch == '\\')
		    state = 1;
		else if (ch == '\'')
		    state = 13;
		else
		    ret += ch;
		break;

	    case 1:
		// After backslash
		
		switch(ch) {
		case 'S':
		    state = 2;
		    break;

		case 'P':
		    state = 3;
		    break;

		case 'X':
		    state = 4;
		    break;

		default:
		    throw new Error ("Illegal escape: "+ch);
		}
		break;

	    case 2:
		// After \S
		if (ch == '\\')
		    state = 5;
		else
		    throw new Error ("Expecting \ after \S");
		break;


	    case 3:
		// After \P
		state = 6;
		break;


	    case 4:
		// After \X
		switch (ch) {
		case '\\':
		    state = 7;
		    break;

		case '2':
		    hex = "";
		    hex_idx = 0;
		    hex_sz = 4;
		    state = 9;
		    break;
		    
		case '4':
		    hex = "";
		    hex_idx = 0;
		    hex_sz = 8;
		    state = 9;
		    break;
		    
		}
		break;

	    case 5:
		// After \S\  Char += 128
		ret += String.fromCharCode(ch.charCodeAt(0) + 128);
		state = 0;
		break;

	    case 6:
		// After \PX.  This should set the code page, but
		// we are not implementing this since it requires translation
		// tables, and nobody really uses this anyway
		if (ch == '\\')
		    state = 0;
		else
		    throw new Error ("Expecting \\ after \\Px");
		
		break;

	    case 7:
		// After \X\  get 2 hex digits
		hex = ch;
		state = 8;
		break;

	    case 8:
		// After \X\h
		assert_hex(ch);
		hex += ch;
		state = 0;
		ret += String.fromCharCode(parseInt(hex, 16));
		break;
		
	    case 9:
		if (ch == '\\') {
		    state = 10;
		}
		else {
		    assert_hex(ch);
		    hex += ch;
		    if (hex.length >= hex_sz) {
			ret += String.fromCharCode(parseInt(hex, 16));
			hex = "";
		    }
		}
		break;

	    case 10:
		if (ch == 'X')
		    state = 11;
		else
		    throw new Error ("Expecting \\X0\\");
		break;
		
	    case 11:
		if (ch == '0')
		    state = 12;
		else
		    throw new Error ("Expecting \\X0\\");
		break;
		
	    case 12:
		if (ch == '\\')
		    state = 0;
		else
		    throw new Error ("Expecting \\X0\\");
		break;

	    case 13:
		state = 0;
		if (ch != '\'')
		    throw new Error ("Expecting double '");
		
		ret += ch;
	    }
	}

	return ret;
	
    }
    
    cls.prototype = SUBTYPE(P21.Wrapper);

    cls.prototype.toString = function()
    {
 	return this.value;
    }
    
    cls.prototype.valueOf = function()
    {
 	return this.value;
    }
    
    cls.prototype.toP21String = function()
    {
 	return "'" + cls.encodeP21String(this.value, false) + "'";
    }
    
    cls.fromP21String = function(s) {
	return new cls(cls.decodeP21String(s));
    }
    
    return cls;
})();



P21.Enum = (function() {

    var cls = function(val) {
	if (val == null)
	    return null;

 	val = val.valueOf();
	
	if (typeof val === 'boolean') {
	    /* Nothing */
	}

	else {
	    val = val.toString();
	    
	    val = val.toUpperCase();

	    var first = val.charAt(0);
	    if ((first < 'A' || first > 'Z') && first != '_')
		throw new Error ("Illegal enum value");
	    
	    if (val == 'T')
		val = true;
	    else if (val == 'F')
		val = false;
 	}
	
 	this.value = val;
    }
    
    cls.prototype = SUBTYPE(P21.Wrapper);
    
    cls.prototype.toString = function()
    {
 	return this.value;
    }
    
    cls.prototype.valueOf = function()
    {
 	return this.value;
    }
    
    cls.prototype.toP21String = function()
    {
	if (this.value == null)
	    return "$";
	
	if (typeof this.value === 'boolean') {
	    if (this.value)
		return '.T.';
	    else
		return '.F.';
	}
	
 	return "." + this.value + ".";
    }
    
     return cls;
})();



P21.Binary = (function() {

    var cls = function(val) {
	if (val == null)
	    return null;

 	val = val.valueOf();

 	if (typeof val !== 'string') {
 	    val = new String(val).toString()
 	}
	
 	this.value = val;
    }

     cls.prototype = SUBTYPE(P21.Wrapper);

     cls.prototype.toString = function()
     {
 	return this.value;
     }
  
     cls.prototype.valueOf = function()
     {
 	return this.value;
     }
  
     cls.prototype.toP21String = function()
     {
 	return "\"" + this.value + "\"";
     }
     
     return cls;
})();


P21.URI = (function() {

    var cls = function(val) {
	
	if (val == null)
	    return null;

 	val = val.valueOf();

 	if (typeof val !== 'string') {
 	    val = new String(val).toString()
 	}
	
 	this.value = val;
    }

     cls.prototype = SUBTYPE(P21.Wrapper);

     cls.prototype.toString = function()
     {
 	return this.value;
     }
  
     cls.prototype.valueOf = function()
     {
 	return this.value;
     }
  
     cls.prototype.toP21String = function()
     {
 	return "<" + this.value + ">";
     }
     
     return cls;
})();


P21.OMITTED = {

    valueOf : function() {return null;},

    toString : function() {return "";},

    toP21String : function() {return "*";}    
};
    

P21.ID = (function() {

    var cls = function(val) {
 	if (typeof val !== 'string') {
 	    val = new val.toString()
 	}
	
	val  = val.replace(/^#/, "");

	if (val.search(/\D/) == -1) {
	    val = val.replace(/^0*/, '');
	}
	
 	this.value = val;
    }
    
    cls.prototype = SUBTYPE(P21.Wrapper);

    cls.prototype.toString = function()
    {
 	return this.value.toString();
    }
    
    cls.prototype.valueOf = function()
    {
 	return this.value;
    }
  
    cls.prototype.toP21String = function()
    {
 	return "#" + this.value;
    }
    
    return cls;
})();
